@@ -16,7 +16,7 @@
  * for each file.  Finally, it sorts the array to the final order that the
  * actions should be executed in.
  *
- * Copyright (c) 2013-2023, PostgreSQL Global Development Group
+ * Copyright (c) 2013-2024, PostgreSQL Global Development Group
  *
  *-------------------------------------------------------------------------
  */
@@ -27,26 +27,25 @@
 #include <unistd.h>
 
 #include "catalog/pg_tablespace_d.h"
-#include "common/hashfn.h"
+#include "common/file_utils.h"
+#include "common/hashfn_unstable.h"
 #include "common/string.h"
 #include "datapagemap.h"
 #include "filemap.h"
 #include "pg_rewind.h"
-#include "storage/fd.h"
 
 /*
  * Define a hash table which we can use to store information about the files
  * appearing in source and target systems.
  */
-static uint32 hash_string_pointer(const char *s);
-#define SH_PREFIX				filehash
-#define SH_ELEMENT_TYPE			file_entry_t
-#define SH_KEY_TYPE				const char *
-#define SH_KEY					path
-#define SH_HASH_KEY(tb, key)	hash_string_pointer(key)
+#define SH_PREFIX		filehash
+#define SH_ELEMENT_TYPE	file_entry_t
+#define SH_KEY_TYPE		const char *
+#define	SH_KEY			path
+#define SH_HASH_KEY(tb, key)	hash_string(key)
 #define SH_EQUAL(tb, a, b)		(strcmp(a, b) == 0)
-#define SH_SCOPE				static inline
-#define SH_RAW_ALLOCATOR		pg_malloc0
+#define	SH_SCOPE		static inline
+#define SH_RAW_ALLOCATOR	pg_malloc0
 #define SH_DECLARE
 #define SH_DEFINE
 #include "lib/simplehash.h"
@@ -61,36 +60,7 @@ static char *datasegpath(RelFileLocator rlocator, ForkNumber forknum,
 
 static file_entry_t *insert_filehash_entry(const char *path);
 static file_entry_t *lookup_filehash_entry(const char *path);
-
-/*
- * A separate hash table which tracks WAL files that must not be deleted.
- */
-typedef struct keepwal_entry
-{
-	const char *path;
-	uint32		status;
-} keepwal_entry;
-
-#define SH_PREFIX				keepwal
-#define SH_ELEMENT_TYPE			keepwal_entry
-#define SH_KEY_TYPE				const char *
-#define SH_KEY					path
-#define SH_HASH_KEY(tb, key)	hash_string_pointer(key)
-#define SH_EQUAL(tb, a, b)		(strcmp(a, b) == 0)
-#define SH_SCOPE				static inline
-#define SH_RAW_ALLOCATOR		pg_malloc0
-#define SH_DECLARE
-#define SH_DEFINE
-#include "lib/simplehash.h"
-
-#define KEEPWAL_INITIAL_SIZE	1000
-
-
-static keepwal_hash *keepwal = NULL;
-static bool keepwal_entry_exists(const char *path);
-
 static int	final_filemap_cmp(const void *a, const void *b);
-
 static bool check_file_excluded(const char *path, bool is_source);
 
 /*
@@ -114,7 +84,7 @@ struct exclude_list_item
  * they are defined in backend-only headers.  So this list is maintained
  * with a best effort in mind.
  */
-static const char *excludeDirContents[] =
+static const char *const excludeDirContents[] =
 {
 	/*
 	 * Skip temporary statistics files. PG_STAT_TMP_DIR must be skipped
@@ -236,39 +206,6 @@ lookup_filehash_entry(const char *path)
 	return filehash_lookup(filehash, path);
 }
 
-/*
- * Initialize a hash table to store WAL file names that must be kept.
- */
-void
-keepwal_init(void)
-{
-	/* An initial hash size out of thin air */
-	keepwal = keepwal_create(KEEPWAL_INITIAL_SIZE, NULL);
-}
-
-/* Mark the given file to prevent its removal */
-void
-keepwal_add_entry(const char *path)
-{
-	keepwal_entry *entry;
-	bool		found;
-
-	/* Should only be called with keepwal initialized */
-	Assert(keepwal != NULL);
-
-	entry = keepwal_insert(keepwal, path, &found);
-
-	if (!found)
-		entry->path = pg_strdup(path);
-}
-
-/* Return true if file is marked as not to be removed, false otherwise */
-static bool
-keepwal_entry_exists(const char *path)
-{
-	return keepwal_lookup(keepwal, path) != NULL;
-}
-
 /*
  * Callback for processing source file list.
  *
@@ -748,15 +685,7 @@ decide_file_action(file_entry_t *entry)
 	}
 	else if (entry->target_exists && !entry->source_exists)
 	{
-		/*
-		 * For files that exist in target but not in source, we check the
-		 * keepwal hash table; any files listed therein must not be removed.
-		 */
-		if (keepwal_entry_exists(path))
-		{
-			pg_log_debug("Not removing file \"%s\" because it is required for recovery", path);
-			return FILE_ACTION_NONE;
-		}
+		/* File exists in target, but not source. Remove it. */
 		return FILE_ACTION_REMOVE;
 	}
 	else if (!entry->target_exists && !entry->source_exists)
@@ -891,15 +820,3 @@ decide_file_actions(void)
 
 	return filemap;
 }
-
-
-/*
- * Helper function for filemap hash table.
- */
-static uint32
-hash_string_pointer(const char *s)
-{
-	unsigned char *ss = (unsigned char *) s;
-
-	return hash_bytes(ss, strlen(s));
-}
